@@ -16,6 +16,9 @@ import re
 from .sysinfo import RUNNING_ON_APPVEYOR as APPVEYOR
 from .sysinfo import RUNNING_ON_TRAVIS as TRAVIS
 from .sysinfo import RESOLVER_NOT_SYSTEM as ARES
+from .sysinfo import RESOLVER_ARES
+from .sysinfo import RESOLVER_DNSPYTHON
+from .sysinfo import RUNNING_ON_CI
 from .sysinfo import RUN_COVERAGE
 
 
@@ -1229,6 +1232,13 @@ if PY38:
             'test_ssl.BasicSocketTests.test_parse_cert_CVE_2013_4238',
         ]
 
+if RESOLVER_DNSPYTHON:
+    disabled_tests += [
+        # This does two things DNS python doesn't. First, it sends it
+        # capital letters and expects them to be returned lowercase.
+        # Second, it expects the symbolic scopeid to be stripped from the end.
+        'test_socket.GeneralModuleTests.test_getaddrinfo_ipv6_scopeid_symbolic',
+    ]
 
 # if 'signalfd' in os.environ.get('GEVENT_BACKEND', ''):
 #     # tests that don't interact well with signalfd
@@ -1254,6 +1264,14 @@ if OSX:
         'test_socket.RecvmsgIntoTCPTest.testRecvmsgIntoGenerator',
     ]
 
+    if RESOLVER_ARES and PY38 and not RUNNING_ON_CI:
+        disabled_tests += [
+            # When updating to 1.16.0 this was seen locally, but not on CI.
+            # Tuples differ: ('ff02::1de:c0:face:8d', 1234, 0, 0)
+            #             != ('ff02::1de:c0:face:8d', 1234, 0, 1)
+            'test_socket.GeneralModuleTests.test_getaddrinfo_ipv6_scopeid_symbolic',
+        ]
+
 # Now build up the data structure we'll use to actually find disabled tests
 # to avoid a linear scan for every file (it seems the list could get quite large)
 # (First, freeze the source list to make sure it isn't modified anywhere)
